@@ -9,11 +9,12 @@ import java.util.Optional;
 public class Branch {
   public BoardData data;
 
-  public Branch(Board board, List<String> variation) {
-    int[] moveNumberList = new int[Board.boardSize * Board.boardSize];
+  public Branch(Board board, List<String> variation, int length) {
+    int[] moveNumberList = new int[Board.boardWidth * Board.boardHeight];
     int moveNumber = 0;
     double winrate = 0.0;
     int playouts = 0;
+    double scoreMean = 0.0;
 
     this.data =
         new BoardData(
@@ -27,11 +28,13 @@ public class Branch {
             board.getData().blackCaptures,
             board.getData().whiteCaptures,
             winrate,
-            playouts);
+            playouts,
+            scoreMean);
 
-    for (int i = 0; i < variation.size(); i++) {
+    length = length > 0 ? length : 361;
+    for (int i = 0; i < variation.size() && i < length; i++) {
       Optional<int[]> coordOpt = Board.asCoordinates(variation.get(i));
-      if (!coordOpt.isPresent()) {
+      if (!coordOpt.isPresent() || !Board.isValid(coordOpt.get()[0], coordOpt.get()[1])) {
         break;
       }
       int[] coord = coordOpt.get();
