@@ -4,7 +4,7 @@ import java.util.Date;
 
 public class GameInfo {
   public static final String DEFAULT_NAME_HUMAN_PLAYER = "Human";
-  public static final String DEFAULT_NAME_CPU_PLAYER = "Leela Zero";
+  public static final String DEFAULT_NAME_CPU_PLAYER = "AI";
   public static final double DEFAULT_KOMI = 7.5;
 
   private String playerBlack = "";
