@@ -17,14 +17,33 @@ import java.util.Map;
 import java.util.Optional;
 import java.util.Queue;
 import java.util.function.BiConsumer;
+import java.util.regex.Matcher;
+import java.util.regex.Pattern;
 import java.util.stream.Collectors;
 import java.util.stream.Stream;
-import javax.swing.*;
+import javax.swing.JOptionPane;
 import org.json.JSONException;
 
 public class Board implements LeelazListener {
-  public static int boardSize = Lizzie.config.config.getJSONObject("ui").optInt("board-size", 19);
-  private static final String alphabet = "ABCDEFGHJKLMNOPQRST";
+  public static int boardSize0 =
+      (Lizzie.config != null
+              && Lizzie.config.config != null
+              && Lizzie.config.config.getJSONObject("ui") != null)
+          ? Lizzie.config.config.getJSONObject("ui").optInt("board-size", 19)
+          : 19;
+  public static int boardWidth =
+      (Lizzie.config != null
+              && Lizzie.config.config != null
+              && Lizzie.config.config.getJSONObject("ui") != null)
+          ? Lizzie.config.config.getJSONObject("ui").optInt("board-width", boardSize0)
+          : boardSize0;
+  public static int boardHeight =
+      (Lizzie.config != null
+              && Lizzie.config.config != null
+              && Lizzie.config.config.getJSONObject("ui") != null)
+          ? Lizzie.config.config.getJSONObject("ui").optInt("board-height", boardSize0)
+          : boardSize0;
+  private static final String alphabet = "ABCDEFGHJKLMNOPQRSTUVWXYZ";
 
   private BoardHistoryList history;
   private Stone[] capturedStones;
@@ -32,12 +51,16 @@ public class Board implements LeelazListener {
   private boolean analysisMode;
   private int playoutsAnalysis;
 
+  public String allowCoords = "";
+  public String avoidCoords = "";
+  public boolean isForcing = false;
+  public boolean isAllowing = false;
+  public boolean isAvoding = false;
+  public boolean isKeepingAvoid = false;
+
   // Save the node for restore move when in the branch
   private Optional<BoardHistoryNode> saveNode;
 
-  // Force refresh board
-  private boolean forceRefresh;
-
   public Board() {
     initialize();
   }
@@ -49,8 +72,8 @@ public class Board implements LeelazListener {
     analysisMode = false;
     playoutsAnalysis = 100;
     saveNode = Optional.empty();
-    forceRefresh = false;
-    history = new BoardHistoryList(BoardData.empty(boardSize));
+    Lizzie.frame.setForceRefresh(false);
+    history = new BoardHistoryList(BoardData.empty(boardWidth, boardHeight));
   }
 
   /**
@@ -61,7 +84,13 @@ public class Board implements LeelazListener {
    * @return the array index
    */
   public static int getIndex(int x, int y) {
-    return x * Board.boardSize + y;
+    return x * Board.boardHeight + y;
+  }
+
+  public static int[] getCoord(int index) {
+    int y = index % Board.boardHeight;
+    int x = (index - y) / Board.boardHeight;
+    return new int[] {x, y};
   }
 
   /**
@@ -77,9 +106,73 @@ public class Board implements LeelazListener {
       return Optional.empty();
     }
     // coordinates take the form C16 A19 Q5 K10 etc. I is not used.
-    int x = alphabet.indexOf(namedCoordinate.charAt(0));
-    int y = boardSize - Integer.parseInt(namedCoordinate.substring(1));
-    return Optional.of(new int[] {x, y});
+    String reg = "([A-HJ-Z]+)(\\d+)";
+    Pattern p = Pattern.compile(reg);
+    Matcher m = p.matcher(namedCoordinate);
+    if (m.find() && m.groupCount() == 2) {
+      int x = asDigit(m.group(1));
+      int y = boardHeight - Integer.parseInt(m.group(2));
+      return Optional.of(new int[] {x, y});
+    } else {
+      reg = "\\(([\\d]+),([\\d]+)\\)";
+      p = Pattern.compile(reg);
+      m = p.matcher(namedCoordinate);
+      if (m.find() && m.groupCount() == 2) {
+        int x = Integer.parseInt(m.group(1));
+        int y = Integer.parseInt(m.group(2)); // boardHeight - Integer.parseInt(m.group(2)) - 1;
+        return Optional.of(new int[] {x, y});
+      } else {
+        return Optional.empty();
+      }
+    }
+  }
+
+  public static int asDigit(String name) {
+    // coordinates take the form C16 A19 Q5 K10 etc. I is not used.
+    int base = alphabet.length();
+    char names[] = name.toCharArray();
+    int length = names.length;
+    if (length > 0) {
+      int x = 0;
+      for (int i = length - 1; i >= 0; i--) {
+        int index = alphabet.indexOf(names[i]);
+        if (index == -1) {
+          return index;
+        }
+        x += index * Math.pow(base, length - i - 1);
+      }
+      return x;
+    } else {
+      return -1;
+    }
+  }
+
+  public static String asName(int c) {
+    return asName(c, true);
+  }
+
+  public void clearAnalysis() {
+    getData().tryToClearBestMoves();
+  }
+
+  public static String asName(int c, boolean isName) {
+    if (boardWidth > 25 && isName) {
+      return String.valueOf(c + 1);
+    }
+    StringBuilder name = new StringBuilder();
+    int base = alphabet.length();
+    int n = c;
+    ArrayDeque<Integer> ad = new ArrayDeque<Integer>();
+    if (n > 0) {
+      while (n > 0) {
+        ad.addFirst(n < 25 && c >= 25 ? n % base - 1 : n % base);
+        n /= base;
+      }
+    } else {
+      ad.addFirst(n);
+    }
+    ad.forEach(i -> name.append(alphabet.charAt(i)));
+    return name.toString();
   }
 
   /**
@@ -91,7 +184,11 @@ public class Board implements LeelazListener {
    */
   public static String convertCoordinatesToName(int x, int y) {
     // coordinates take the form C16 A19 Q5 K10 etc. I is not used.
-    return alphabet.charAt(x) + "" + (boardSize - y);
+    if (boardWidth > 25 || boardHeight > 25) {
+      return String.format("(%d,%d)", x, y); // boardHeight - y - 1);
+    } else {
+      return asName(x, false) + (boardHeight - y);
+    }
   }
 
   /**
@@ -102,7 +199,11 @@ public class Board implements LeelazListener {
    * @return whether or not this coordinate is part of the board
    */
   public static boolean isValid(int x, int y) {
-    return x >= 0 && x < boardSize && y >= 0 && y < boardSize;
+    return x >= 0 && x < boardWidth && y >= 0 && y < boardHeight;
+  }
+
+  public static boolean isValid(int[] c) {
+    return c != null && c.length == 2 && isValid(c[0], c[1]);
   }
 
   /**
@@ -110,23 +211,19 @@ public class Board implements LeelazListener {
    *
    * @param size
    */
-  public void reopen(int size) {
-    size = (size == 13 || size == 9) ? size : 19;
-    if (size != boardSize) {
-      boardSize = size;
-      initialize();
-      forceRefresh = true;
+  public void reopen(int width, int height) {
+    width = (width >= 2) ? width : 19;
+    height = (height >= 2) ? height : 19;
+    if (width != boardWidth || height != boardHeight) {
+      boardWidth = width;
+      boardHeight = height;
+      Zobrist.init();
+      clear();
+      Lizzie.leelaz.boardSize(boardWidth, boardHeight);
+      Lizzie.frame.setForceRefresh(true);
     }
   }
 
-  public boolean isForceRefresh() {
-    return forceRefresh;
-  }
-
-  public void setForceRefresh(boolean forceRefresh) {
-    this.forceRefresh = forceRefresh;
-  }
-
   /**
    * The comment. Thread safe
    *
@@ -163,6 +260,37 @@ public class Board implements LeelazListener {
     }
   }
 
+  public int moveNumberByCoord(int[] coord) {
+    int moveNumber = 0;
+    if (Lizzie.board.isValid(coord)) {
+      int index = Lizzie.board.getIndex(coord[0], coord[1]);
+      if (Lizzie.board.getHistory().getStones()[index] != Stone.EMPTY) {
+        BoardHistoryNode cur = Lizzie.board.getHistory().getCurrentHistoryNode();
+        moveNumber = cur.getData().moveNumberList[index];
+        if (!cur.isMainTrunk()) {
+          if (moveNumber > 0) {
+            moveNumber = cur.getData().moveNumber - cur.getData().moveMNNumber + moveNumber;
+          } else {
+            BoardHistoryNode p = cur.firstParentWithVariations().orElse(cur);
+            while (p != cur && moveNumber == 0) {
+              moveNumber = p.getData().moveNumberList[index];
+              if (moveNumber > 0) {
+                BoardHistoryNode topOfTop = p.firstParentWithVariations().orElse(p);
+                if (topOfTop != p) {
+                  moveNumber = p.getData().moveNumber - p.getData().moveMNNumber + moveNumber;
+                }
+              } else {
+                cur = p;
+                p = cur.firstParentWithVariations().orElse(cur);
+              }
+            }
+          }
+        }
+      }
+    }
+    return moveNumber;
+  }
+
   /**
    * Add a stone to the board representation. Thread safe
    *
@@ -181,7 +309,7 @@ public class Board implements LeelazListener {
       stones[getIndex(x, y)] = color;
       zobrist.toggleStone(x, y, color);
 
-      Lizzie.frame.repaint();
+      Lizzie.frame.refresh();
     }
   }
 
@@ -206,7 +334,7 @@ public class Board implements LeelazListener {
       zobrist.toggleStone(x, y, oriColor);
       data.moveNumberList[Board.getIndex(x, y)] = 0;
 
-      Lizzie.frame.repaint();
+      Lizzie.frame.refresh();
     }
   }
 
@@ -228,7 +356,7 @@ public class Board implements LeelazListener {
   /**
    * Add a keys and values to node
    *
-   * @param map
+   * @param properties
    */
   public void addNodeProperties(Map<String, String> properties) {
     synchronized (this) {
@@ -236,13 +364,26 @@ public class Board implements LeelazListener {
     }
   }
 
+  public void setKomi(double komi) {
+    getHistory().getGameInfo().setKomi(komi);
+    Lizzie.leelaz.komi(komi);
+  }
+
   /**
    * The pass. Thread safe
    *
    * @param color the type of pass
    */
   public void pass(Stone color) {
-    pass(color, false, false);
+    pass(color, false, false, false);
+  }
+
+  public void pass(Stone color, boolean newBranch) {
+    pass(color, newBranch, false, false);
+  }
+
+  public void pass(Stone color, boolean newBranch, boolean dummy) {
+    pass(color, newBranch, dummy, false);
   }
 
   /**
@@ -251,7 +392,7 @@ public class Board implements LeelazListener {
    * @param color the type of pass
    * @param newBranch add a new branch
    */
-  public void pass(Stone color, boolean newBranch, boolean dummy) {
+  public void pass(Stone color, boolean newBranch, boolean dummy, boolean changeMove) {
     synchronized (this) {
 
       // check to see if this move is being replayed in history
@@ -270,8 +411,8 @@ public class Board implements LeelazListener {
       Zobrist zobrist = history.getZobrist();
       int moveNumber = history.getMoveNumber() + 1;
       int[] moveNumberList =
-          newBranch && history.getNext().isPresent()
-              ? new int[Board.boardSize * Board.boardSize]
+          newBranch && history.getNext(true).isPresent()
+              ? new int[Board.boardWidth * Board.boardHeight]
               : history.getMoveNumberList().clone();
 
       // build the new game state
@@ -286,19 +427,20 @@ public class Board implements LeelazListener {
               moveNumberList,
               history.getData().blackCaptures,
               history.getData().whiteCaptures,
+              0.0,
               0,
-              0);
+              0.0);
       newState.dummy = dummy;
 
       // update leelaz with pass
-      Lizzie.leelaz.playMove(color, "pass");
+      if (!Lizzie.leelaz.isInputCommand) Lizzie.leelaz.playMove(color, "pass");
       if (Lizzie.frame.isPlayingAgainstLeelaz)
         Lizzie.leelaz.genmove((history.isBlacksTurn() ? "W" : "B"));
 
       // update history with pass
-      history.addOrGoto(newState, newBranch);
+      history.addOrGoto(newState, newBranch, changeMove);
 
-      Lizzie.frame.repaint();
+      Lizzie.frame.refresh();
     }
   }
 
@@ -318,6 +460,10 @@ public class Board implements LeelazListener {
     place(x, y, color, false);
   }
 
+  public void place(int x, int y, Stone color, boolean newBranch) {
+    place(x, y, color, newBranch, false);
+  }
+
   /**
    * Places a stone onto the board representation. Thread safe
    *
@@ -326,7 +472,8 @@ public class Board implements LeelazListener {
    * @param color the type of stone to place
    * @param newBranch add a new branch
    */
-  public void place(int x, int y, Stone color, boolean newBranch) {
+  public void place(int x, int y, Stone color, boolean newBranch, boolean changeMove) {
+    Lizzie.frame.clearBeforeMove();
     synchronized (this) {
       if (scoreMode) {
         // Mark clicked stone as dead
@@ -341,12 +488,18 @@ public class Board implements LeelazListener {
       updateWinrate();
       double nextWinrate = -100;
       if (history.getData().winrate >= 0) nextWinrate = 100 - history.getData().winrate;
+      double nextScoreMean = -history.getData().scoreMean;
 
       // check to see if this coordinate is being replayed in history
       Optional<int[]> nextLast = history.getNext().flatMap(n -> n.lastMove);
-      if (nextLast.isPresent() && nextLast.get()[0] == x && nextLast.get()[1] == y && !newBranch) {
+      if (nextLast.isPresent()
+          && nextLast.get()[0] == x
+          && nextLast.get()[1] == y
+          && !newBranch
+          && !changeMove) {
         // this is the next coordinate in history. Just increment history so that we don't erase the
         // redo's
+        Lizzie.leelaz.beginModifyingBoard();
         history.next();
         // should be opposite from the bottom case
         if (Lizzie.frame.isPlayingAgainstLeelaz
@@ -356,6 +509,7 @@ public class Board implements LeelazListener {
         } else if (!Lizzie.frame.isPlayingAgainstLeelaz) {
           Lizzie.leelaz.playMove(color, convertCoordinatesToName(x, y));
         }
+        Lizzie.leelaz.endModifyingBoard();
         return;
       }
 
@@ -367,8 +521,8 @@ public class Board implements LeelazListener {
       int moveMNNumber =
           history.getMoveMNNumber() > -1 && !newBranch ? history.getMoveMNNumber() + 1 : -1;
       int[] moveNumberList =
-          newBranch && history.getNext().isPresent()
-              ? new int[Board.boardSize * Board.boardSize]
+          newBranch && history.getNext(true).isPresent()
+              ? new int[Board.boardWidth * Board.boardHeight]
               : history.getMoveNumberList().clone();
 
       moveNumberList[Board.getIndex(x, y)] = moveMNNumber > -1 ? moveMNNumber : moveNumber;
@@ -387,7 +541,7 @@ public class Board implements LeelazListener {
       // check to see if the player made a suicidal coordinate
       int isSuicidal = removeDeadChain(x, y, color, stones, zobrist);
 
-      for (int i = 0; i < boardSize * boardSize; i++) {
+      for (int i = 0; i < boardWidth * boardHeight; i++) {
         if (stones[i].equals(Stone.EMPTY)) {
           moveNumberList[i] = 0;
         }
@@ -409,25 +563,29 @@ public class Board implements LeelazListener {
               bc,
               wc,
               nextWinrate,
-              0);
+              0,
+              nextScoreMean);
       newState.moveMNNumber = moveMNNumber;
+      newState.dummy = false;
 
       // don't make this coordinate if it is suicidal or violates superko
       if (isSuicidal > 0 || history.violatesKoRule(newState)) return;
 
       // update leelaz with board position
+      Lizzie.leelaz.beginModifyingBoard();
       if (Lizzie.frame.isPlayingAgainstLeelaz
           && Lizzie.frame.playerIsBlack == getData().blackToPlay) {
         Lizzie.leelaz.playMove(color, convertCoordinatesToName(x, y));
         Lizzie.leelaz.genmove((Lizzie.board.getData().blackToPlay ? "W" : "B"));
-      } else if (!Lizzie.frame.isPlayingAgainstLeelaz) {
+      } else if (!Lizzie.frame.isPlayingAgainstLeelaz && !Lizzie.leelaz.isInputCommand) {
         Lizzie.leelaz.playMove(color, convertCoordinatesToName(x, y));
       }
 
       // update history with this coordinate
-      history.addOrGoto(newState, newBranch);
+      history.addOrGoto(newState, newBranch, changeMove);
+      Lizzie.leelaz.endModifyingBoard();
 
-      Lizzie.frame.repaint();
+      Lizzie.frame.refresh();
     }
   }
 
@@ -471,11 +629,12 @@ public class Board implements LeelazListener {
                 blackToPlay,
                 zobrist,
                 0,
-                new int[boardSize * boardSize],
+                new int[boardWidth * boardHeight],
                 0,
                 0,
                 0.0,
-                0));
+                0,
+                0.0));
     history.setGameInfo(oldHistory.getGameInfo());
   }
 
@@ -489,7 +648,7 @@ public class Board implements LeelazListener {
    * @param zobrist the zobrist object to modify
    * @return number of removed stones
    */
-  private int removeDeadChain(int x, int y, Stone color, Stone[] stones, Zobrist zobrist) {
+  public static int removeDeadChain(int x, int y, Stone color, Stone[] stones, Zobrist zobrist) {
     if (!isValid(x, y) || stones[getIndex(x, y)] != color) return 0;
 
     boolean hasLiberties = hasLibertiesHelper(x, y, color, stones);
@@ -508,7 +667,7 @@ public class Board implements LeelazListener {
    * @param stones the stones array to modify
    * @return whether or not this chain has liberties
    */
-  private boolean hasLibertiesHelper(int x, int y, Stone color, Stone[] stones) {
+  private static boolean hasLibertiesHelper(int x, int y, Stone color, Stone[] stones) {
     if (!isValid(x, y)) return false;
 
     if (stones[getIndex(x, y)] == Stone.EMPTY) return true; // a liberty was found
@@ -540,7 +699,7 @@ public class Board implements LeelazListener {
    *     their unrecursed version
    * @return number of removed stones
    */
-  private int cleanupHasLibertiesHelper(
+  private static int cleanupHasLibertiesHelper(
       int x, int y, Stone color, Stone[] stones, Zobrist zobrist, boolean removeStones) {
     int removed = 0;
     if (!isValid(x, y) || stones[getIndex(x, y)] != color) return 0;
@@ -597,8 +756,10 @@ public class Board implements LeelazListener {
 
   /** Goes to the next coordinate, thread safe */
   public boolean nextMove() {
+    Lizzie.frame.clearBeforeMove();
     synchronized (this) {
       updateWinrate();
+      Lizzie.leelaz.beginModifyingBoard();
       if (history.next().isPresent()) {
         // update leelaz board position, before updating to next node
         Optional<int[]> lastMoveOpt = history.getData().lastMove;
@@ -609,9 +770,11 @@ public class Board implements LeelazListener {
         } else {
           Lizzie.leelaz.playMove(history.getLastMoveColor(), "pass");
         }
-        Lizzie.frame.repaint();
+        Lizzie.frame.refresh();
+        Lizzie.leelaz.endModifyingBoard();
         return true;
       }
+      Lizzie.leelaz.endModifyingBoard();
       return false;
     }
   }
@@ -646,8 +809,11 @@ public class Board implements LeelazListener {
   /** Save the back routing from children */
   public void saveBackRouting(BoardHistoryNode node) {
     Optional<BoardHistoryNode> prev = node.previous();
-    prev.ifPresent(n -> n.setFromBackChildren(n.getVariations().indexOf(node)));
-    prev.ifPresent(n -> n.previous().ifPresent(p -> saveBackRouting(p)));
+    prev.ifPresent(
+        n -> {
+          n.setFromBackChildren(n.getVariations().indexOf(node));
+          saveBackRouting(n);
+        });
   }
 
   /** Restore move number by saved node */
@@ -661,8 +827,8 @@ public class Board implements LeelazListener {
     for (int i = 0; i < stones.length; i++) {
       Stone stone = stones[i];
       if (stone.isBlack() || stone.isWhite()) {
-        int y = i % Board.boardSize;
-        int x = (i - y) / Board.boardSize;
+        int y = i % Board.boardWidth;
+        int x = (i - y) / Board.boardHeight;
         Lizzie.leelaz.playMove(stone, convertCoordinatesToName(x, y));
       }
     }
@@ -740,7 +906,7 @@ public class Board implements LeelazListener {
         } else {
           Lizzie.leelaz.playMove(history.getLastMoveColor(), "pass");
         }
-        Lizzie.frame.repaint();
+        Lizzie.frame.refresh();
         return true;
       }
       return false;
@@ -896,7 +1062,7 @@ public class Board implements LeelazListener {
   public void deleteMove() {
     synchronized (this) {
       BoardHistoryNode currentNode = history.getCurrentHistoryNode();
-      if (currentNode.next().isPresent()) {
+      if (currentNode.next(true).isPresent()) {
         // Will delete more than one move, ask for confirmation
         int ret =
             JOptionPane.showConfirmDialog(
@@ -940,22 +1106,27 @@ public class Board implements LeelazListener {
 
   /** Clears all history and starts over from empty board. */
   public void clear() {
-    Lizzie.leelaz.sendCommand("clear_board");
+    Lizzie.leelaz.clear();
     Lizzie.frame.resetTitle();
     Lizzie.frame.clear();
     initialize();
+    setKomi(getHistory().getGameInfo().getKomi());
   }
 
   /** Goes to the previous coordinate, thread safe */
   public boolean previousMove() {
+    Lizzie.frame.clearBeforeMove();
     synchronized (this) {
       if (inScoreMode()) setScoreMode(false);
       updateWinrate();
+      Lizzie.leelaz.beginModifyingBoard();
       if (history.previous().isPresent()) {
         Lizzie.leelaz.undo();
-        Lizzie.frame.repaint();
+        Lizzie.frame.refresh();
+        Lizzie.leelaz.endModifyingBoard();
         return true;
       }
+      Lizzie.leelaz.endModifyingBoard();
       return false;
     }
   }
@@ -1019,7 +1190,7 @@ public class Board implements LeelazListener {
 
       lastup = lastdown = false;
       // Find all stones within empty area line by line (new lines added to Q)
-      while (x < boardSize) {
+      while (x < boardWidth) {
         if (shdwstones[getIndex(x, y)] == Stone.EMPTY) {
           shdwstones[getIndex(x, y)] = Stone.DAME; // Too mark that it has been visited
         } else if (stones[getIndex(x, y)] == toggle) {
@@ -1037,7 +1208,7 @@ public class Board implements LeelazListener {
           lastup = false;
         }
         // Check below
-        if (y + 1 < boardSize
+        if (y + 1 < boardHeight
             && (shdwstones[getIndex(x, y + 1)] == Stone.EMPTY
                 || stones[getIndex(x, y + 1)] == toggle)) {
           if (!lastdown) visitQ.add(new int[] {x, y + 1});
@@ -1048,7 +1219,7 @@ public class Board implements LeelazListener {
         x++;
       }
     }
-    Lizzie.frame.repaint();
+    Lizzie.frame.refresh();
   }
 
   /*
@@ -1095,7 +1266,7 @@ public class Board implements LeelazListener {
       }
 
       lastup = lastdown = false;
-      while (x < boardSize && emptyOrCaptured(shdwstones, x, y)) {
+      while (x < boardWidth && emptyOrCaptured(shdwstones, x, y)) {
         // Check above
         if (y - 1 >= 0 && shdwstones[getIndex(x, y - 1)] != Stone.DAME) {
           if (emptyOrCaptured(shdwstones, x, y - 1)) {
@@ -1113,7 +1284,7 @@ public class Board implements LeelazListener {
           }
         }
         // Check below
-        if (y + 1 < boardSize && shdwstones[getIndex(x, y + 1)] != Stone.DAME) {
+        if (y + 1 < boardHeight && shdwstones[getIndex(x, y + 1)] != Stone.DAME) {
           if (emptyOrCaptured(shdwstones, x, y + 1)) {
             if (!lastdown) {
               visitQ.add(new int[] {x, y + 1});
@@ -1138,7 +1309,7 @@ public class Board implements LeelazListener {
         x++;
       }
       // At this point x is at the edge of the board or on a stone
-      if (x < boardSize && shdwstones[getIndex(x, y)] != found) {
+      if (x < boardWidth && shdwstones[getIndex(x, y)] != found) {
         if (found == Stone.EMPTY) found = shdwstones[getIndex(x, y)];
         else found = Stone.DAME;
       }
@@ -1161,8 +1332,8 @@ public class Board implements LeelazListener {
 
     Stone[] scoreStones = capturedStones.clone();
 
-    for (int i = 0; i < boardSize; i++) {
-      for (int j = 0; j < boardSize; j++) {
+    for (int i = 0; i < boardWidth; i++) {
+      for (int j = 0; j < boardHeight; j++) {
         if (scoreStones[getIndex(i, j)] == Stone.EMPTY) {
           markEmptyArea(scoreStones, i, j);
         }
@@ -1179,8 +1350,8 @@ public class Board implements LeelazListener {
         new double[] {
           getData().blackCaptures, getData().whiteCaptures + getHistory().getGameInfo().getKomi()
         };
-    for (int i = 0; i < boardSize; i++) {
-      for (int j = 0; j < boardSize; j++) {
+    for (int i = 0; i < boardWidth; i++) {
+      for (int j = 0; j < boardHeight; j++) {
         switch (scoreStones[getIndex(i, j)]) {
           case BLACK_POINT:
             score[0]++;
@@ -1241,6 +1412,7 @@ public class Board implements LeelazListener {
       if (!history.getNext().isPresent() || isSuccessivePass) {
         // Reached the end...
         toggleAnalysis();
+        Lizzie.leelaz.togglePonder();
       } else if (bestMoves.isEmpty()) {
         // If we get empty list, something strange happened, ignore notification
       } else {
@@ -1288,9 +1460,202 @@ public class Board implements LeelazListener {
 
   public void updateWinrate() {
     Leelaz.WinrateStats stats = Lizzie.leelaz.getWinrateStats();
-    if (stats.maxWinrate >= 0 && stats.totalPlayouts > history.getData().playouts) {
+    if (stats.maxWinrate >= 0 && stats.totalPlayouts > history.getData().getPlayouts()) {
       history.getData().winrate = stats.maxWinrate;
-      history.getData().playouts = stats.totalPlayouts;
+      // we won't set playouts here. but setting winrate is ok... it shows the user that we are
+      // computing. i think its fine.
     }
   }
+
+  public boolean changeMove(int moveNumber, String changeMove) {
+    Optional<int[]> changeCoord = asCoordinates(changeMove);
+    if ("pass".equalsIgnoreCase(changeMove)) {
+      changeMove(moveNumber, (int[]) null);
+      return true;
+    } else if ("swap".equalsIgnoreCase(changeMove)) {
+      changeMove(moveNumber, null, true);
+      return true;
+    } else if (changeCoord.isPresent()
+        && Board.isValid(changeCoord.get()[0], changeCoord.get()[1])) {
+      changeCoord.map(c -> changeMove(moveNumber, c, false));
+      return true;
+    } else {
+      return false;
+    }
+  }
+
+  public boolean changeMove(int moveNumber, int[] coords) {
+    return changeMove(moveNumber, coords, false);
+  }
+
+  public boolean changeMove(int moveNumber, int[] coords, boolean swapColorOnly) {
+    if (moveNumber <= 0) {
+      return false;
+    }
+
+    int endMoveNumber = history.getEnd().moveNumberOfNode();
+    if (moveNumber > endMoveNumber) {
+      return false;
+    }
+
+    int currentMoveNumber = history.getMoveNumber();
+
+    goToMoveNumber(moveNumber - 1);
+
+    Optional<BoardHistoryNode> changeNode = history.getCurrentHistoryNode().next();
+    Optional<BoardHistoryNode> relink = changeNode.flatMap(n -> n.next());
+
+    // Change Move
+    if (swapColorOnly) {
+      if (changeNode.isPresent()) {
+        Optional<int[]> c = changeNode.get().getData().lastMove;
+        if (c.isPresent() && isValid(c.get())) {
+          changeNode
+              .map(n -> n.getData())
+              .map(d -> d.lastMoveColor)
+              .ifPresent(s -> place(c.get()[0], c.get()[1], s.opposite(), false, true));
+        } else {
+          pass(history.isBlacksTurn() ? Stone.BLACK : Stone.WHITE, false, false, true);
+        }
+      }
+    } else {
+      if (coords != null && Board.isValid(coords[0], coords[1])) {
+        place(
+            coords[0], coords[1], history.isBlacksTurn() ? Stone.BLACK : Stone.WHITE, false, true);
+      } else {
+        pass(history.isBlacksTurn() ? Stone.BLACK : Stone.WHITE, false, false, true);
+      }
+    }
+
+    Optional<BoardHistoryNode> node = relink;
+    while (node.isPresent()) {
+      Optional<int[]> lastMove = node.get().getData().lastMove;
+      if (lastMove.isPresent()) {
+        int[] m = lastMove.get();
+        if (Board.isValid(m[0], m[1])) {
+          place(
+              m[0],
+              m[1],
+              swapColorOnly
+                  ? node.get().getData().lastMoveColor
+                  : history.isBlacksTurn() ? Stone.BLACK : Stone.WHITE,
+              false,
+              true);
+        } else {
+          pass(
+              swapColorOnly
+                  ? node.get().getData().lastMoveColor
+                  : history.isBlacksTurn() ? Stone.BLACK : Stone.WHITE,
+              false,
+              false,
+              true);
+        }
+        node = node.get().next();
+      }
+    }
+
+    goToMoveNumber(currentMoveNumber);
+
+    return true;
+  }
+
+  public ArrayList<MoveList> getMoveList() {
+    ArrayList<MoveList> moveList = new ArrayList<MoveList>();
+
+    Optional<BoardHistoryNode> node = history.getCurrentHistoryNode().now();
+    Optional<int[]> passstep = Optional.empty();
+    while (node.isPresent()) {
+      Optional<int[]> lastMove = node.get().getData().lastMove;
+      if (lastMove == passstep) {
+        MoveList move = new MoveList();
+        move.isPass = true;
+        move.isBlack = node.get().getData().lastMoveColor.isBlack();
+        moveList.add(move);
+        node = node.get().previous();
+      } else {
+        if (lastMove.isPresent()) {
+
+          int[] n = lastMove.get();
+          MoveList move = new MoveList();
+          move.x = n[0];
+          move.y = n[1];
+          move.isPass = false;
+          move.isBlack = node.get().getData().lastMoveColor.isBlack();
+          move.moveNum = node.get().getData().moveNumber;
+          moveList.add(move);
+          node = node.get().previous();
+        }
+      }
+    }
+    moveList.remove(moveList.size() - 1);
+    return moveList;
+  }
+
+  public boolean setAvoidCoords(int x, int y) {
+    Optional<int[]> boardCoordinates = Lizzie.frame.convertScreenToCoordinates(x, y);
+    if (boardCoordinates.isPresent()) {
+      if (isCoordsEmpty(boardCoordinates.get()[0], boardCoordinates.get()[1])) {
+        if (avoidCoords == "")
+          avoidCoords =
+              convertCoordinatesToName(boardCoordinates.get()[0], boardCoordinates.get()[1]);
+        else
+          avoidCoords =
+              avoidCoords
+                  + ","
+                  + convertCoordinatesToName(boardCoordinates.get()[0], boardCoordinates.get()[1]);
+        return true;
+      } else return false;
+    }
+    return false;
+  }
+
+  public boolean setAllowCoords(int x, int y) {
+    Optional<int[]> boardCoordinates = Lizzie.frame.convertScreenToCoordinates(x, y);
+    if (boardCoordinates.isPresent()) {
+      if (isCoordsEmpty(boardCoordinates.get()[0], boardCoordinates.get()[1])) {
+        allowCoords =
+            convertCoordinatesToName(boardCoordinates.get()[0], boardCoordinates.get()[1]);
+        return true;
+      } else return false;
+    }
+    return false;
+  }
+
+  public boolean addAllowCoords(int x, int y) {
+    Optional<int[]> boardCoordinates = Lizzie.frame.convertScreenToCoordinates(x, y);
+    if (boardCoordinates.isPresent()) {
+      if (isCoordsEmpty(boardCoordinates.get()[0], boardCoordinates.get()[1])) {
+        allowCoords =
+            allowCoords
+                + ","
+                + convertCoordinatesToName(boardCoordinates.get()[0], boardCoordinates.get()[1]);
+        return true;
+      } else return false;
+    }
+    return false;
+  }
+
+  public boolean isCoordsEmpty(int x, int y) {
+    if (history.getStones()[getIndex(x, y)] != Stone.EMPTY) {
+      return false;
+    }
+    return true;
+  }
+
+  public boolean setAsMainBranch() {
+    if (history.getCurrentHistoryNode().isMainTrunk()) return false;
+    BoardHistoryNode topNode = history.getCurrentHistoryNode().topOfFatherBranch();
+    BoardHistoryNode mainNode = history.getCurrentHistoryNode().nodeBeforeTopOfFatherBranch();
+    BoardHistoryNode oldFirstVar = mainNode.variations.get(0);
+    for (int i = 0; i < mainNode.variations.size(); i++) {
+      if (mainNode.variations.get(i) == topNode) {
+        mainNode.variations.remove(i);
+        mainNode.variations.add(i, oldFirstVar);
+        mainNode.variations.remove(0);
+        mainNode.variations.add(0, topNode);
+        return true;
+      }
+    }
+    return false;
+  }
 }
